#include<cstdio>
#include<iomanip>
#include<string.h>
#include"config.h"
#include"gettext.h"
#include "RtMidi.h"
#include "alsa/asoundef.h"
#define MIDI_CMD_COMMON MIDI_CMD_COMMON_SYSEX
#define _(String) gettext(String)

class MidiMonitor
{
private: 
  static char * keynames[12];
  unsigned char lastcommand;
public:
  typedef std::vector<unsigned char> datamessage;

  MidiMonitor();
  ~MidiMonitor();

  static void MidiCallback( double deltatime, datamessage *message, void *userData );
  static void printNoteOffEvent (unsigned char d1, unsigned  char d2, unsigned int d3);
  static void printNoteOnEvent (unsigned char d1, unsigned char d2, unsigned int d3);
  static void printAfterTouchEvent(unsigned char d1, unsigned char d2, unsigned int d3);
  static void printCtrlChangeEvent(unsigned char d1, unsigned int d2, unsigned int d3);
  static void printProgramChangeEvent(unsigned char d1, unsigned int d2);
  static void printChannelPressureEvent(unsigned char d1, unsigned int d2);
  static void printPitchChangeEvent(unsigned char d1, unsigned int d2, unsigned int d3);
  static int printSystemEvent(datamessage *message, int position);
 
protected:
  RtMidiIn *midiin;
  RtMidiOut *midiout;
};

char* MidiMonitor::keynames[12]=
  {"c","cis/des","d","dis/es","e",
   "f","fis/ges","g","gis/as","a",
   "ais/b","h"};

MidiMonitor::MidiMonitor():lastcommand(0)
{

  try {
    midiin = new RtMidiIn();
  }
  catch (RtError &error) {
    // Handle the exception here
    error.printMessage();
  }
  try {
    midiout = new RtMidiOut();
  }
  catch (RtError &error) {
    // Handle the exception here
    error.printMessage();
  }
  midiin->setCallback(&MidiMonitor::MidiCallback,this);
  midiin->ignoreTypes(false,false,false);
  midiout->openVirtualPort();
  midiin->openVirtualPort();
}

MidiMonitor::~MidiMonitor() 
{
  delete midiin;
  delete midiout;
}

void MidiMonitor::MidiCallback( double deltatime, datamessage *message, void *userData )
{
  datamessage::size_type nBytes = message->size();
  
  if (nBytes >1 && message->at(0) != MIDI_CMD_COMMON_SENSING &&
      message->at(0) & 0x80)
    std::cout << std::endl << "-------------------------------";
  for ( datamessage::size_type i=0; i<nBytes; i++ ){
    if (message->at(i)&0x80) {
      (static_cast<MidiMonitor *>(userData)->lastcommand)=message->at(i);
    }
    else if (static_cast<MidiMonitor *>(userData)->lastcommand) {
      message->insert(message->begin()+i,
		      static_cast<MidiMonitor *>(userData)->lastcommand);
      nBytes = message->size();
    }
    switch ((int)message->at(i) & MIDI_CMD_COMMON) {
    case MIDI_CMD_NOTE_OFF: 
      if (nBytes-i>=2) {
	printNoteOffEvent(message->at(i),message->at(i+1),message->at(i+2)); 
	i+=2; //zwei Bytes zusätzlich
      } else {
	std::cout << _("Bogus Note Off: ") 
		  << std::hex << (int)message->at(i) << std::dec << " ";
      }
      break;
    case MIDI_CMD_NOTE_ON:
      if (nBytes-i>=2) {
	printNoteOnEvent(message->at(i),message->at(i+1),message->at(i+2)); 
	i+=2; //zwei Bytes zusätzlich
      } else {
	std::cout << _("Bogus Note On: ") 
		  << std::hex << (int)message->at(i) << std::dec << " "; 
      }
      break;
    case MIDI_CMD_NOTE_PRESSURE:
      if (nBytes-i>=2) {
	printAfterTouchEvent(message->at(i),message->at(i+1),message->at(i+2)); 
	i+=2; //zwei Bytes zusätzlich
      } else {
	std::cout << _("Bogus AfterTouch: ")
		  << std::hex << (int)message->at(i) << std::dec << " "; 
      }
      break;
    case MIDI_CMD_CONTROL:
      if (nBytes-i>=2) {
	printCtrlChangeEvent(message->at(i),message->at(i+1),message->at(i+2)); 
	i+=2; //zwei Bytes zusätzlich
      } else {
	std::cout << _("Bogus Control Change: ") 
		  << std::hex << (int)message->at(i) << std::dec << " "; 
      }
      break;
    case MIDI_CMD_PGM_CHANGE:
      if (nBytes-i>=1) {
	printProgramChangeEvent(message->at(i),message->at(i+1)); 
	i++; //ein Byte zusätzlich
      } else {
	std::cout << _("Bogus Program Change: ") 
		  << std::hex << (int)message->at(i) << std::dec << " "; 
      }
      break;
    case MIDI_CMD_CHANNEL_PRESSURE:
      if (nBytes-i>=2) {
	printChannelPressureEvent(message->at(i),message->at(i+1)); 
	i++; //ein Byte zusätzlich
      } else {
	std::cout << _("Bogus Program Change: ") 
		  << std::hex << (int)message->at(i) << std::dec << " "; 
      }
      break;
    case MIDI_CMD_BENDER:
      if (nBytes-i>=2) {
	printPitchChangeEvent(message->at(i),message->at(i+1),message->at(i+2)); 
	i+=2; //zwei Bytes zusätzlich
      } else {
	std::cout << _("Bogus Control Change: ") 
		  << std::hex << (int)message->at(i) << std::dec << " "; 
      }
      break;
    case MIDI_CMD_COMMON:
      i += printSystemEvent(message,i);
      break;
      
    default:
      std::cout << _("Byte ") << i << " = " 
		<< std::hex << (int)message->at(i) << std::dec << ",\t";
    }
  }
  if ( nBytes > 0 && message->at(nBytes-1)!=MIDI_CMD_COMMON_SENSING 
       && message->at(nBytes-1)!=MIDI_CMD_COMMON_CLOCK)
    std::cout << _(" stamp = ") << deltatime << '\n';
    static_cast<MidiMonitor *>(userData)->midiout->sendMessage(message);
}

void MidiMonitor::printNoteOffEvent(unsigned char d1, unsigned char d2, unsigned int d3)
{
  std::cout << std::endl << _("Off\tChannel: ") 
	    << std::setw(2) << (d1 & 0xf) << _("\t Key: ")
	    << (keynames[d2 % 12]) 
	    << std::setw(2) << (d2/12-6) << _("\t Velocity: ") 
	    << std::setw(3) << d3 <<" ";
}
void MidiMonitor::printNoteOnEvent(unsigned char d1, unsigned char d2, unsigned int d3)
{
  std::cout << std::endl << _("On\tChannel: ") 
	    << std::setw(2) << (d1 & 0xf) << _("\t Key: ")
	    << (keynames[d2 % 12]) 
	    << std::setw(2) << (d2/12-6) << _("\t Velocity: ") 
	    << std::setw(3) << d3 << " ";
}
void MidiMonitor::printAfterTouchEvent(unsigned char d1, unsigned char d2, unsigned int d3)
{
  std::cout << std::endl << _("AftTch\tChannel: ") 
	    << std::setw(2) << (d1 & 0xf) << _("\t Key: ")
	    << (keynames[d2 % 12]) 
	    << std::setw(2) << (d2/12-6) << _("\t Velocity: ") 
	    << std::setw(3) << d3 << " ";
}
void MidiMonitor::printCtrlChangeEvent(unsigned char d1, unsigned int d2, unsigned int d3)
{
  if (d2<120) {
    std::cout << std::endl << _("CtrlChg\tChannel: ") 
	      << std::setw(2) << (d1 & 0xf) 
	      << _("  Ctrl: ") << std::setw(3)
	      << d2 << "  ";
    switch(d2) {
    case MIDI_CTL_MSB_BANK: std::cout << _("Bank MSB"); break;
    case MIDI_CTL_MSB_MODWHEEL: std::cout << _("ModW MSB"); break;
    case MIDI_CTL_MSB_BREATH: std::cout << _("Breath MSB"); break;
    case MIDI_CTL_MSB_FOOT: std::cout << _("Foot MSB"); break;
    case MIDI_CTL_MSB_PORTAMENTO_TIME: std::cout << _("PortTime MSB"); break;
    case MIDI_CTL_MSB_DATA_ENTRY: std::cout << _("DataEntry MSB"); break;
    case MIDI_CTL_MSB_MAIN_VOLUME: std::cout << _("MainVol MSB"); break;
    case MIDI_CTL_MSB_BALANCE: std::cout << _("Balance MSB"); break;
    case MIDI_CTL_MSB_PAN: std::cout << _("Pan MSB"); break;
    case MIDI_CTL_MSB_EXPRESSION: std::cout << _("Express MSB"); break;
    case MIDI_CTL_MSB_EFFECT1: std::cout << _("Eff 1 MSB"); break;
    case MIDI_CTL_MSB_EFFECT2: std::cout << _("Eff 2 MSB"); break;
    case MIDI_CTL_MSB_GENERAL_PURPOSE1: std::cout << _("GenPurp1 MSB"); break;
    case MIDI_CTL_MSB_GENERAL_PURPOSE2: std::cout << _("GenPurp2 MSB"); break;
    case MIDI_CTL_MSB_GENERAL_PURPOSE3: std::cout << _("GenPurp3 MSB"); break;
    case MIDI_CTL_MSB_GENERAL_PURPOSE4: std::cout << _("GenPurp4 MSB"); break;
    case MIDI_CTL_LSB_BANK: std::cout << _("Bank LSB"); break;
    case MIDI_CTL_LSB_MODWHEEL: std::cout << _("ModW LSB"); break;
    case MIDI_CTL_LSB_BREATH: std::cout << _("Breath LSB"); break;
    case MIDI_CTL_LSB_FOOT: std::cout << _("Foot LSB"); break;
    case MIDI_CTL_LSB_PORTAMENTO_TIME: std::cout << _("PortTime LSB"); break;
    case MIDI_CTL_LSB_DATA_ENTRY: std::cout << _("DataEntry LSB"); break;
    case MIDI_CTL_LSB_MAIN_VOLUME: std::cout << _("MainVol LSB"); break;
    case MIDI_CTL_LSB_BALANCE: std::cout << _("Balance LSB"); break;
    case MIDI_CTL_LSB_PAN: std::cout << _("Pan LSB"); break;
    case MIDI_CTL_LSB_EXPRESSION: std::cout << _("Express LSB"); break;
    case MIDI_CTL_LSB_EFFECT1: std::cout << _("Eff 1 LSB"); break;
    case MIDI_CTL_LSB_EFFECT2: std::cout << _("Eff 2 LSB"); break;
    case MIDI_CTL_LSB_GENERAL_PURPOSE1: std::cout << _("GenPurp1 LSB"); break;
    case MIDI_CTL_LSB_GENERAL_PURPOSE2: std::cout << _("GenPurp2 LSB"); break;
    case MIDI_CTL_LSB_GENERAL_PURPOSE3: std::cout << _("GenPurp3 LSB"); break;
    case MIDI_CTL_LSB_GENERAL_PURPOSE4: std::cout << _("GenPurp4 LSB"); break;
    case MIDI_CTL_SUSTAIN: std::cout << _("Sustain"); break;
    case MIDI_CTL_PORTAMENTO: std::cout << _("Portamento"); break;
    case MIDI_CTL_SUSTENUTO: std::cout << _("Sustenuto"); break;
    case MIDI_CTL_SOFT_PEDAL: std::cout << _("Soft pedal"); break;
    case MIDI_CTL_LEGATO_FOOTSWITCH: std::cout << _("Legato Footswitch"); break;
    case MIDI_CTL_HOLD2: std::cout << _("Hold 2"); break;
    case MIDI_CTL_SC1_SOUND_VARIATION: std::cout << _("SC1 SoundVar"); break;
    case MIDI_CTL_SC2_TIMBRE: std::cout << _("SC2 Timbre"); break;
    case MIDI_CTL_SC3_RELEASE_TIME: std::cout << _("SC3 RelTime"); break;
    case MIDI_CTL_SC4_ATTACK_TIME: std::cout << _("SC4 AttackTime"); break;
    case MIDI_CTL_SC5_BRIGHTNESS: std::cout << _("SC5 Brightness"); break;
    case MIDI_CTL_SC6: std::cout << _("SC6 "); break;
    case MIDI_CTL_SC7: std::cout << _("SC7 "); break;
    case MIDI_CTL_SC8: std::cout << _("SC8 "); break;
    case MIDI_CTL_SC9: std::cout << _("SC9 "); break;
    case MIDI_CTL_SC10: std::cout << _("SC10"); break;
    case MIDI_CTL_GENERAL_PURPOSE5: std::cout << _("GenPurp5"); break;
    case MIDI_CTL_GENERAL_PURPOSE6: std::cout << _("GenPurp6"); break;
    case MIDI_CTL_GENERAL_PURPOSE7: std::cout << _("GenPurp7"); break;
    case MIDI_CTL_GENERAL_PURPOSE8: std::cout << _("GenPurp8"); break;
    case MIDI_CTL_PORTAMENTO_CONTROL: std::cout << _("Portamento Control"); break;
    case MIDI_CTL_E1_REVERB_DEPTH: std::cout << _("E1 Reverb Depth"); break;
    case MIDI_CTL_E2_TREMOLO_DEPTH: std::cout << _("E2 Tremolo Depth"); break;
    case MIDI_CTL_E3_CHORUS_DEPTH: std::cout << _("E3 Chorus Depth"); break;
    case MIDI_CTL_E4_DETUNE_DEPTH: std::cout << _("E4 Detune Depth"); break;
    case MIDI_CTL_E5_PHASER_DEPTH: std::cout << _("E5 Phaser Depth"); break;
    case MIDI_CTL_DATA_INCREMENT: std::cout << _("Data Increment"); break;
    case MIDI_CTL_DATA_DECREMENT: std::cout << _("Data Decrement"); break;
    case MIDI_CTL_NONREG_PARM_NUM_LSB: std::cout << _("NonReg LSB"); break;
    case MIDI_CTL_NONREG_PARM_NUM_MSB: std::cout << _("NonReg MSB"); break;
    case MIDI_CTL_REGIST_PARM_NUM_LSB: std::cout << _("Regist LSB"); break;
    case MIDI_CTL_REGIST_PARM_NUM_MSB: std::cout << _("Regist MSB"); break;
    }
    std::cout << _("\tValue: ") << std::setw(3) << d3 << " ";
  } else {
    std::cout << std::endl << _("CommSys\tChannel: ") << (d1 & 0xf)
	      << _("  Ctrl: ") << std::setw(3)
	      << d2 << "  ";
    switch(d2) {
    case MIDI_CTL_ALL_SOUNDS_OFF: std::cout << _("All Sound off"); break;
    case MIDI_CTL_RESET_CONTROLLERS: std::cout << _("Reset Controllers"); break;
    case MIDI_CTL_LOCAL_CONTROL_SWITCH: std::cout << _("Local ContSwitch"); break;
    case MIDI_CTL_ALL_NOTES_OFF: std::cout << _("All Notes Off"); break;
    case MIDI_CTL_OMNI_OFF: std::cout << _("Omni Off"); break;
    case MIDI_CTL_OMNI_ON: std::cout << _("Omni On"); break;
    case MIDI_CTL_MONO1: std::cout << _("Mono1"); break;
    case MIDI_CTL_MONO2: std::cout << _("Mono2"); break;
    }
    std::cout  << _("\tValue: ") << std::setw(3) << d3 << " ";
  }
}
void MidiMonitor::printProgramChangeEvent(unsigned char d1, unsigned int d2)
{
  std::cout << std::endl << _("ProgChg\tChannel: ") 
	    << std::setw(2) << (d1 & 0xf) << _("\t Program: ")
	    << std::setw(3) << d2  << " ";
}
void MidiMonitor::printChannelPressureEvent(unsigned char d1, unsigned int d2)
{
  std::cout << std::endl << _("ChaPres\tChannel: ") << (d1 & 0xf) << _("\t Velocity: ")
	    << d2  << " ";
}
void MidiMonitor::printPitchChangeEvent(unsigned char d1, unsigned int d2, unsigned int d3)
{
  std::cout << std::endl << _("PitchCh\tChannel: ") 
	    << std::setw(2) << (d1 & 0xf) << _("\t Value: ")
	    << std::setw(5)
	    << (signed int)(d2+(d3 << 7))- (signed int)0x2000 << " ";
}
int MidiMonitor::printSystemEvent(datamessage *message, int position)
{
  unsigned int nBytes=message->size(), i=position;
  switch (message->at(i)) {
  case MIDI_CMD_COMMON_SYSEX: 
    std::cout << std::endl << _("System exclusive, Vendor=") << std::hex <<
      (int)message->at(i+1) << std::dec << std::endl;
     for (i+=2; i<nBytes && message->at(i)!=MIDI_CMD_COMMON_SYSEX_END;i++) {
       std::cout << std::hex << std::setw(2) << std::setfill('0')
		 << (int)message->at(i) << std::dec << " ";
       if ((i-position)%8==1) std::cout << " ";
       if ((i-position)%16==1) 
	 std::cout << "\t" << nBytes << " " << i << std::endl;
     }
     std::cout << std::hex << std::setw(2) << std::setfill('0')
	       << (int)message->at(i) << std::dec << " ";
     if ((i-position)%8==1) std::cout << " ";
       std::cout << "\t" << nBytes << " " << i << std::endl;
     if (i<nBytes){
       if (message->at(i)==MIDI_CMD_COMMON_SYSEX_END) 
	 std::cout << std::endl << _("System exclusive end.");
       else std::cout << std::endl << _("Unfinished system exclusive.");
     }
       else std::cout << std::endl << _("Unfinished system exclusive.");
     std::cout << std::flush << std::setfill(' ');
    return i-position;
  case MIDI_CMD_COMMON_MTC_QUARTER:
    std::cout << std::endl << _("??? MTC_QUARTER.");
    return 0;
  case MIDI_CMD_COMMON_SONG_POS: 
    std::cout << std::endl << _("Song position Pointer: ") 
              << std::setw(6)
	      << message->at(i+1) + message->at(i+2)<<7;
    return 2;
  case MIDI_CMD_COMMON_SONG_SELECT: 
    std::cout << std::endl << _("Song Select: ") 
	      << std::setw(3) << (int)message->at(i+1);
    return 1;
  case 0xf4:
  case 0xf5:
    std::cout << std::endl << _("Undefined System Message (") 
	      << std::hex << message->at(i) << std::dec << ").";
    return 0;
  case MIDI_CMD_COMMON_TUNE_REQUEST:
    std::cout << std::endl << _("Tune Request.");
    return 0;
  case MIDI_CMD_COMMON_SYSEX_END:
    std::cout << std::endl << _("Unexpected SysEx end!");
    return 0;
  case MIDI_CMD_COMMON_CLOCK:
    std::cout << "+" << std::flush;
    return 0;
  case MIDI_CMD_COMMON_START:
    std::cout << std::endl << _("Sequencer start.") ;
    return 0;
  case MIDI_CMD_COMMON_CONTINUE:
    std::cout << std::endl << _("Sequencer continue.") ;
    return 0;
  case MIDI_CMD_COMMON_STOP:
    std::cout << std::endl << _("Sequencer stop.");
    return 0;
  case MIDI_CMD_COMMON_SENSING:
    (std::cout << ".").flush();
    return 0;
  case MIDI_CMD_COMMON_RESET:
    std::cout << std::endl << _("Reset.");
    return 0;
  default:
    std::cout << _(" Byte ") << std::hex << std::setw(2) << std::setfill('0')
	      << int (message->at(i)) << std::dec  << std::setfill(' ')
	      << "\t";
  }
  return 0;
}


MidiMonitor Monitor;

int main()
{
  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);
  std::cerr << PACKAGE_STRING 
	    << _( " monitoring MIDI data in clear text.") 
	    << std::endl << std::endl;

  std::cerr << _("    Copyright (C) 2005  Tobias Schlemmer\n"
		 "    This program is free software; you can redistribute it and/or modify\n"
		 "    it under the terms of the GNU General Public License as published by\n"
		 "    the Free Software Foundation; either version 2 of the License, or\n"
		 "    (at your option) any later version.\n\n"

		 "    This program is distributed in the hope that it will be useful,\n"
		 "    but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
		 "    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
		 "    GNU General Public License for more details.\n\n"
		 
		 "    You should have received a copy of the GNU General Public License\n"
		 "    along with this program; if not, write to the Free Software\n"
		 "    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA" )
	    << std::endl << std::endl;

  std::cerr << _("Waiting for data.") << std::endl;
  while (getc(stdin)!='e') ;
}
